==> fred-inquisitor-1.0.0rc2.tar.gz: fred-inquisitor-1.0.0rc2/inquisitor/auth.py <==
"""Auth endpoints."""
from datetime import datetime, timedelta
from typing import cast

from fastapi import APIRouter, Depends, HTTPException, Request, Response, Security, status
from fastapi.openapi.models import OAuthFlows as OAuthFlowsModel
from fastapi.param_functions import Form
from fastapi.security import OAuth2, OAuth2PasswordRequestForm, SecurityScopes
from fastapi.security.utils import get_authorization_scheme_param
from fred_types import ContactId
from grill import Logger
from jose import JWTError, jwe, jwt
from jose.constants import ALGORITHMS
from jose.exceptions import JWEError
from pydantic import BaseModel
from verifier import Verification, VerificationClient, VerificationId
from verifier.exceptions import VerificationDoesNotExist

from .constants import (
    COMMON_RESPONSES,
    OAUTH_COOKIE,
    LogEntryType,
    LoggerResult,
    LogReferenceType,
    Scope,
)
from .models import Error
from .settings import Settings, get_logger, get_settings, get_verification_client
from .utils import get_source_ip


class OAuth2PasswordBearerWithCookie(OAuth2):
    """OAuth bearer stored in cookie."""

    def __init__(self, *, tokenUrl: str, scopes: dict, scheme_name: str | None = None):
        flows = OAuthFlowsModel(password={"tokenUrl": tokenUrl, "scopes": scopes})
        super().__init__(flows=flows, scheme_name=scheme_name)

    async def __call__(self, request: Request) -> str:
        """Get access token from cookie."""
        authorization: str | None = request.cookies.get(OAUTH_COOKIE)

        scheme, param = get_authorization_scheme_param(authorization)
        if not authorization or scheme.lower() != "bearer":
            raise HTTPException(
                status_code=status.HTTP_401_UNAUTHORIZED,
                detail="Not authenticated",
                headers={"WWW-Authenticate": "Bearer"},
            )

        return param


oauth2_scheme = OAuth2PasswordBearerWithCookie(
    tokenUrl="auth/token/",
    scopes={
        Scope.READ: "Read verification data",
        Scope.WRITE: "Perform operations with verification.",
    },
)

router = APIRouter(prefix="/auth")


class TokenData(BaseModel):
    """Auth token data."""

    verification_id: VerificationId
    contact_id: ContactId
    scopes: list[Scope] = []


class OAuth2RequestForm(OAuth2PasswordRequestForm):
    """OAuth2 request form with optional password field."""

    def __init__(
        self,
        grant_type: str = Form(default=None, regex="password"),
        username: str = Form(),
        password: str = Form(default=None),
        scope: str = Form(default=""),
        client_id: str | None = Form(default=None),
        client_secret: str | None = Form(default=None),
    ):
        super().__init__(
            grant_type=grant_type,
            username=username,
            password=password,
            scope=scope,
            client_id=client_id,
            client_secret=client_secret,
        )


def create_access_token(data: TokenData, settings: Settings) -> str:
    """Create access JWT."""
    to_encode = data.dict()
    to_encode.update(
        {
            # issued at
            "iat": datetime.utcnow(),
            # not valid before
            "nbf": datetime.utcnow() - timedelta(minutes=5),
            # expiration time
            "exp": datetime.utcnow() + timedelta(seconds=settings.access_token_expires),
        }
    )
    encoded_jwt = jwt.encode(to_encode, settings.secret_key, algorithm=ALGORITHMS.HS256)
    encrypted_jwt = jwe.encrypt(
        encoded_jwt,
        key=settings.secret_key_sha256,
        algorithm=ALGORITHMS.DIR,
        encryption=ALGORITHMS.A256GCM,
    ).decode()
    return cast(str, encrypted_jwt)


async def auth_token_data(
    security_scopes: SecurityScopes,
    token: str = Depends(oauth2_scheme),
    settings: Settings = Depends(get_settings),
) -> TokenData:
    """Get token data from encrypted JWT."""
    try:
        decrypted_jwt = jwe.decrypt(token, settings.secret_key_sha256)
        payload = jwt.decode(decrypted_jwt, settings.secret_key, algorithms=[ALGORITHMS.HS256])
        data = TokenData(**payload)

        # Check security scopes
        if set(security_scopes.scopes) <= set(data.scopes):
            return data
        else:
            raise HTTPException(
                status_code=status.HTTP_401_UNAUTHORIZED,
                detail="Insufficient permissions",
                headers={"WWW-Authenticate": f'Bearer scope="{security_scopes.scope_str}"'},
            )
    except (JWTError, JWEError):
        raise HTTPException(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail="Could not validate credentials",
            headers={"WWW-Authenticate": "Bearer"},
        )


async def get_contact_id(
    data: TokenData = Security(auth_token_data, scopes=[Scope.READ]),
) -> ContactId:
    """Get contact id."""
    return data.contact_id


async def get_verification_id(
    data: TokenData = Security(auth_token_data, scopes=[Scope.READ]),
) -> VerificationId:
    """Get verification id."""
    return data.verification_id


async def get_verification(
    verification_id: VerificationId = Security(get_verification_id, scopes=[Scope.READ]),
    verification_client: VerificationClient = Depends(get_verification_client),
) -> Verification:
    """Get verification."""
    try:
        return await verification_client.get(verification_id)
    except VerificationDoesNotExist:
        raise HTTPException(status_code=status.HTTP_404_NOT_FOUND, detail="Verification not found")


class LoginResponse(BaseModel):
    """Login response."""

    scopes: list[Scope]


@router.post(
    "/token/",
    responses={
        400: {"description": "Incorrect username or password", "model": Error},
        401: {"description": "Insufficient authorization for requested scopes", "model": Error},
    },
)
async def login(
    request: Request,
    response: Response,
    form_data: OAuth2RequestForm = Depends(),
    source_ip: str | None = Depends(get_source_ip),
    settings: Settings = Depends(get_settings),
    verification_client: VerificationClient = Depends(get_verification_client),
    logger: Logger = Depends(get_logger),
) -> LoginResponse:
    """Log in using OAuth2."""
    log_entry = logger.create(
        LogEntryType.LOGIN,
        source_ip=source_ip,
        properties={
            "username": form_data.username,
            "password": str(bool(form_data.password)),
            "scopes": form_data.scopes,
        },
    )
    with log_entry:
        try:
            verification = await verification_client.get(
                VerificationId(form_data.username), access_token=form_data.password
            )
            log_entry.references = {
                LogReferenceType.VERIFICATION: verification.verification_id,
                LogReferenceType.CONTACT: verification.contact_id,
            }
        except VerificationDoesNotExist:
            log_entry.result = LoggerResult.FAIL
            raise HTTPException(
                status_code=status.HTTP_400_BAD_REQUEST, detail="Incorrect username or password"
            )

        granted_scopes = [Scope.READ]
        if form_data.password:
            granted_scopes.append(Scope.WRITE)
        log_entry.properties = {"granted_scopes": granted_scopes}

        if not set(form_data.scopes) <= set(granted_scopes):
            log_entry.result = LoggerResult.FAIL
            raise HTTPException(
                status_code=status.HTTP_401_UNAUTHORIZED,
                detail="Insufficient authorization for requested scopes",
            )

        access_token = create_access_token(
            TokenData(
                verification_id=verification.verification_id,
                contact_id=verification.contact_id,
                scopes=granted_scopes,
            ),
            settings=settings,
        )

        log_entry.result = LoggerResult.SUCCESS
        response.set_cookie(
            key=OAUTH_COOKIE,
            value=f"Bearer {access_token}",
            secure=True,
            httponly=True,
            samesite="lax",
            expires=settings.access_token_expires,
        )
        return LoginResponse(
            scopes=granted_scopes,
        )


@router.post("/token/refresh/", responses=COMMON_RESPONSES)
async def refresh_access_token(
    response: Response,
    token: TokenData = Depends(auth_token_data),
    settings: Settings = Depends(get_settings),
) -> LoginResponse:
    """Refresh access token."""
    access_token = create_access_token(
        TokenData(
            verification_id=token.verification_id,
            contact_id=token.contact_id,
            scopes=token.scopes,
        ),
        settings=settings,
    )

    response.set_cookie(
        key=OAUTH_COOKIE,
        value=f"Bearer {access_token}",
        secure=True,
        httponly=True,
        samesite="none",
        expires=settings.access_token_expires,
    )
    return LoginResponse(
        scopes=token.scopes,
    )

==> fred-inquisitor-1.2.0rc1.tar.gz: fred-inquisitor-1.2.0rc1/inquisitor/auth.py <==
"""Auth endpoints."""
from datetime import datetime, timedelta
from typing import cast

from fastapi import APIRouter, Depends, HTTPException, Request, Response, Security, status
from fastapi.openapi.models import OAuthFlows as OAuthFlowsModel
from fastapi.param_functions import Form
from fastapi.security import OAuth2, OAuth2PasswordRequestForm, SecurityScopes
from fastapi.security.utils import get_authorization_scheme_param
from fred_types import ContactId
from grill import Logger
from jose import JWTError, jwe, jwt
from jose.constants import ALGORITHMS
from jose.exceptions import JWEError
from pydantic import BaseModel
from regal import ContactClient
from regal.exceptions import ContactDoesNotExist
from verifier import Verification, VerificationClient, VerificationId
from verifier.exceptions import VerificationDoesNotExist

from .constants import (
    COMMON_RESPONSES,
    OAUTH_COOKIE,
    LogEntryType,
    LoggerResult,
    LogReferenceType,
    Scope,
)
from .models import Error
from .settings import (
    Settings,
    get_contact_client,
    get_logger,
    get_settings,
    get_verification_client,
)
from .utils import get_source_ip


class OAuth2PasswordBearerWithCookie(OAuth2):
    """OAuth bearer stored in cookie."""

    def __init__(self, *, tokenUrl: str, scopes: dict, scheme_name: str | None = None):
        flows = OAuthFlowsModel(password={"tokenUrl": tokenUrl, "scopes": scopes})
        super().__init__(flows=flows, scheme_name=scheme_name)

    async def __call__(self, request: Request) -> str:
        """Get access token from cookie."""
        authorization: str | None = request.cookies.get(OAUTH_COOKIE)

        scheme, param = get_authorization_scheme_param(authorization)
        if not authorization or scheme.lower() != "bearer":
            raise HTTPException(
                status_code=status.HTTP_401_UNAUTHORIZED,
                detail="Not authenticated",
                headers={"WWW-Authenticate": "Bearer"},
            )

        return param


oauth2_scheme = OAuth2PasswordBearerWithCookie(
    tokenUrl="auth/token/",
    scopes={
        Scope.READ: "Read verification data",
        Scope.WRITE: "Perform operations with verification.",
    },
)

router = APIRouter(prefix="/auth")


class TokenData(BaseModel):
    """Auth token data."""

    verification_id: VerificationId
    contact_id: ContactId
    scopes: list[Scope] = []


class OAuth2RequestForm(OAuth2PasswordRequestForm):
    """OAuth2 request form with optional password field."""

    def __init__(
        self,
        grant_type: str = Form(default=None, regex="password"),
        username: str = Form(),
        password: str = Form(default=None),
        scope: str = Form(default=""),
        client_id: str | None = Form(default=None),
        client_secret: str | None = Form(default=None),
    ):
        super().__init__(
            grant_type=grant_type,
            username=username,
            password=password,
            scope=scope,
            client_id=client_id,
            client_secret=client_secret,
        )


def create_access_token(data: TokenData, settings: Settings) -> str:
    """Create access JWT."""
    to_encode = data.dict()
    to_encode.update(
        {
            # issued at
            "iat": datetime.utcnow(),
            # not valid before
            "nbf": datetime.utcnow() - timedelta(minutes=5),
            # expiration time
            "exp": datetime.utcnow() + timedelta(seconds=settings.access_token_expires),
        }
    )
    encoded_jwt = jwt.encode(to_encode, settings.secret_key, algorithm=ALGORITHMS.HS256)
    encrypted_jwt = jwe.encrypt(
        encoded_jwt,
        key=settings.secret_key_sha256,
        algorithm=ALGORITHMS.DIR,
        encryption=ALGORITHMS.A256GCM,
    ).decode()
    return cast(str, encrypted_jwt)


async def auth_token_data(
    security_scopes: SecurityScopes,
    token: str = Depends(oauth2_scheme),
    settings: Settings = Depends(get_settings),
) -> TokenData:
    """Get token data from encrypted JWT."""
    try:
        decrypted_jwt = jwe.decrypt(token, settings.secret_key_sha256)
        payload = jwt.decode(decrypted_jwt, settings.secret_key, algorithms=[ALGORITHMS.HS256])
        data = TokenData(**payload)

        # Check security scopes
        if set(security_scopes.scopes) <= set(data.scopes):
            return data
        else:
            raise HTTPException(
                status_code=status.HTTP_403_FORBIDDEN,
                detail="Insufficient permissions",
                headers={"WWW-Authenticate": f'Bearer scope="{security_scopes.scope_str}"'},
            )
    except (JWTError, JWEError):
        raise HTTPException(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail="Could not validate credentials",
            headers={"WWW-Authenticate": "Bearer"},
        )


async def get_contact_id(
    data: TokenData = Security(auth_token_data, scopes=[Scope.READ]),
) -> ContactId:
    """Get contact id."""
    return data.contact_id


async def get_verification_id(
    data: TokenData = Security(auth_token_data, scopes=[Scope.READ]),
) -> VerificationId:
    """Get verification id."""
    return data.verification_id


async def get_verification(
    verification_id: VerificationId = Security(get_verification_id, scopes=[Scope.READ]),
    verification_client: VerificationClient = Depends(get_verification_client),
) -> Verification:
    """Get verification."""
    try:
        return await verification_client.get(verification_id)
    except VerificationDoesNotExist:
        raise HTTPException(status_code=status.HTTP_404_NOT_FOUND, detail="Verification not found")


class LoginResponse(BaseModel):
    """Login response."""

    scopes: list[Scope]


async def login(
    *,
    response: Response,
    form_data: OAuth2RequestForm,
    source_ip: str | None,
    settings: Settings,
    verification_client: VerificationClient,
    logger: Logger,
) -> LoginResponse:
    """Log in using OAuth2."""
    log_entry = logger.create(
        LogEntryType.LOGIN,
        source_ip=source_ip,
        properties={
            "username": form_data.username,
            "password": str(bool(form_data.password)),
            "scopes": form_data.scopes,
        },
    )
    with log_entry:
        try:
            verification = await verification_client.get(
                VerificationId(form_data.username), access_token=form_data.password
            )
            log_entry.references = {
                LogReferenceType.VERIFICATION: verification.verification_id,
                LogReferenceType.CONTACT: verification.contact_id,
            }
        except VerificationDoesNotExist:
            log_entry.result = LoggerResult.FAIL
            raise HTTPException(
                status_code=status.HTTP_400_BAD_REQUEST, detail="Incorrect username or password"
            )

        granted_scopes = [Scope.READ]
        if form_data.password:
            granted_scopes.append(Scope.WRITE)
        log_entry.properties = {"granted_scopes": granted_scopes}

        if not set(form_data.scopes) <= set(granted_scopes):
            log_entry.result = LoggerResult.FAIL
            raise HTTPException(
                status_code=status.HTTP_401_UNAUTHORIZED,
                detail="Insufficient authorization for requested scopes",
            )

        access_token = create_access_token(
            TokenData(
                verification_id=verification.verification_id,
                contact_id=verification.contact_id,
                scopes=granted_scopes,
            ),
            settings=settings,
        )

        log_entry.result = LoggerResult.SUCCESS
        response.set_cookie(
            key=OAUTH_COOKIE,
            value=f"Bearer {access_token}",
            secure=True,
            httponly=True,
            samesite="lax",
            expires=settings.access_token_expires,
        )
        return LoginResponse(
            scopes=granted_scopes,
        )


@router.post(
    "/token/",
    responses={
        400: {"description": "Incorrect username or password", "model": Error},
        401: {"description": "Insufficient authorization for requested scopes", "model": Error},
    },
)
async def get_access_token(
    response: Response,
    form_data: OAuth2RequestForm = Depends(),
    source_ip: str | None = Depends(get_source_ip),
    settings: Settings = Depends(get_settings),
    verification_client: VerificationClient = Depends(get_verification_client),
    logger: Logger = Depends(get_logger),
) -> LoginResponse:
    """Log in using OAuth2."""
    # Check whether user supplied password
    if not form_data.password:
        raise HTTPException(status_code=status.HTTP_400_BAD_REQUEST, detail="Password is required")

    return await login(
        response=response,
        form_data=form_data,
        source_ip=source_ip,
        settings=settings,
        verification_client=verification_client,
        logger=logger,
    )


@router.post(
    "/representative/token/",
    responses={
        400: {"description": "Incorrect username or password", "model": Error},
        401: {"description": "Insufficient authorization for requested scopes", "model": Error},
        404: {"description": "Verification not found", "model": Error},
    },
)
async def get_representative_access_token(
    response: Response,
    form_data: OAuth2RequestForm = Depends(),
    source_ip: str | None = Depends(get_source_ip),
    settings: Settings = Depends(get_settings),
    verification_client: VerificationClient = Depends(get_verification_client),
    contact_client: ContactClient = Depends(get_contact_client),
    logger: Logger = Depends(get_logger),
) -> LoginResponse:
    """Log in using OAuth2."""
    # Check whether user supplied password
    if not form_data.password:
        raise HTTPException(status_code=status.HTTP_400_BAD_REQUEST, detail="Password is required")

    handle = form_data.username
    verify_token = form_data.password

    try:
        contact_id = await contact_client.get_contact_id(handle)
    except ContactDoesNotExist:
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST, detail="Incorrect username or password"
        )

    verifications = await verification_client.list(contact_id=contact_id)

    if not verifications.results:
        raise HTTPException(status_code=status.HTTP_404_NOT_FOUND, detail="Verification not found")

    # Find last verification
    verification = max(verifications.results, key=(lambda v: v.create_datetime))

    # Check whether supplied verify_token is valid
    is_valid = await verification_client.verify_token(
        verification.verification_id, verify_token, check=True
    )
    if not is_valid:
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST, detail="Incorrect username or password"
        )

    return await login(
        response=response,
        form_data=OAuth2RequestForm(  # nosec: B106 (hardcoded_password_funcarg)
            username=verification.verification_id, password="", scope=Scope.READ.value
        ),
        source_ip=source_ip,
        settings=settings,
        verification_client=verification_client,
        logger=logger,
    )


@router.post("/token/refresh/", responses=COMMON_RESPONSES)
async def refresh_access_token(
    response: Response,
    token: TokenData = Depends(auth_token_data),
    settings: Settings = Depends(get_settings),
) -> LoginResponse:
    """Refresh access token."""
    access_token = create_access_token(
        TokenData(
            verification_id=token.verification_id,
            contact_id=token.contact_id,
            scopes=token.scopes,
        ),
        settings=settings,
    )

    response.set_cookie(
        key=OAUTH_COOKIE,
        value=f"Bearer {access_token}",
        secure=True,
        httponly=True,
        samesite="none",
        expires=settings.access_token_expires,
    )
    return LoginResponse(
        scopes=token.scopes,
    )


@router.post("/logout/")
async def logout(response: Response) -> None:
    """Logout user (delete access token cookie)."""
    response.delete_cookie(key=OAUTH_COOKIE)
